/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user;

import static java.util.function.Predicate.not;
import static org.assertj.core.api.Assertions.*;

import java.time.Instant;
import java.util.Arrays;
import java.util.List;
import java.util.UUID;
import java.util.stream.Stream;

import jakarta.inject.Inject;

import org.bson.types.ObjectId;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import de.ozgcloud.user.common.MongoDbTestProfile;
import io.quarkus.test.junit.QuarkusTest;
import io.quarkus.test.junit.TestProfile;

@QuarkusTest
@TestProfile(MongoDbTestProfile.class)
class UserRepositoryITCase {

	@Inject
	private UserRepository repository;

	@DisplayName("Update unsynced users")
	@Nested
	class TestUpdateUnsyncedUser {

		private static final long NOW = Instant.now().toEpochMilli();

		@BeforeEach
		void init() {
			repository.deleteAll();
		}

		@Test
		void shouldMarkUnsyncedUserAsDeleted() {
			repository.persist(createUserWithLastSyncTimestamp(NOW - 1000));

			repository.updateUnsyncedUsers(NOW);

			assertThat(getFirstUser().isDeleted()).isTrue();
		}

		@Test
		void shouldNotMarkUserIfSynced() {
			repository.persist(createUserWithLastSyncTimestamp(NOW));

			repository.updateUnsyncedUsers(NOW);

			assertThat(getFirstUser().isDeleted()).isFalse();
		}

		private User createUserWithLastSyncTimestamp(long lastSyncTimestamp) {
			return UserTestFactory.createBuilder().lastSyncTimestamp(lastSyncTimestamp).build();
		}

		private User getFirstUser() {
			return repository.findAll().list().get(0);
		}
	}

	@DisplayName("Find all active user")
	@Nested
	class TestFindAllActiveUser {

		@BeforeEach
		void init() {
			repository.deleteAll();
			repository.persist(UserTestFactory.create());
		}

		@Test
		void shouldFindByLastName() {
			var user = findAllActiveUser(UserTestFactory.LAST_NAME);

			assertThat(user).isNotEmpty().hasSize(1);
		}

		@Test
		void shouldFindCaseInsensitive() {
			var user = findAllActiveUser(UserTestFactory.LAST_NAME.toLowerCase());

			assertThat(user).isNotEmpty().hasSize(1);
		}

		@Test
		void shouldFindByFirstName() {
			var user = findAllActiveUser(UserTestFactory.FIRST_NAME);

			assertThat(user).isNotEmpty().hasSize(1);
		}

		@Test
		void shouldFindByUsername() {
			var user = findAllActiveUser(UserTestFactory.USER_NAME);

			assertThat(user).isNotEmpty().hasSize(1);
		}

		@Test
		void shouldFindByUsernamePart() {
			var user = findAllActiveUser(UserTestFactory.USER_NAME.substring(0, 4));

			assertThat(user).isNotEmpty().hasSize(1);
		}

		@Test
		void shouldFindByEmail() {
			var user = findAllActiveUser(UserTestFactory.EMAIL);

			assertThat(user).isNotEmpty().hasSize(1);
		}

		@Test
		void shouldReturnEmptyListIfNotMatch() {
			var user = findAllActiveUser("xxx");

			assertThat(user).isEmpty();
		}

		private Stream<User> findAllActiveUser(String query) {
			return repository.findUsersByDeleted(query, false, 5);
		}

		@Test
		void shouldFindNotDeletedUsers() {
			User user1 = UserTestFactory.createBuilder().id(new ObjectId()).firstName("Max").deleted(true).build();
			User user2 = UserTestFactory.createBuilder().id(new ObjectId()).firstName("Maximilian").deleted(true).build();
			User user3 = UserTestFactory.createBuilder().id(new ObjectId()).firstName("Maximus").deleted(false).build();

			repository.persist(user1);
			repository.persist(user2);
			repository.persist(user3);

			var foundUsers = findAllActiveUser("Max").toList();

			foundUsers.forEach(u -> assertThat(u.isDeleted()).isFalse());
		}

		@Test
		void shouldUseLimit() {
			repository.persist(UserTestFactory.createBuilder().id(new ObjectId()).build());

			var res = repository.findUsersByDeleted(UserTestFactory.LAST_NAME.toLowerCase(), false, 2);
			assertThat(res).isNotEmpty().hasSize(2);
		}
	}

	@DisplayName("Find all")
	@Nested
	class TestFindAll {

		@BeforeEach
		void init() {
			repository.deleteAll();
			repository.persist(UserTestFactory.createBuilder().id(new ObjectId()).email("test@test.de").externalId("additional-id").build());
			repository.persist(UserTestFactory.create());
		}

		@Test
		void shouldFindDeletedAndNotDeletedUsers() {
			User user1 = UserTestFactory.createBuilder().id(new ObjectId()).firstName("Max").deleted(true).build();
			User user2 = UserTestFactory.createBuilder().id(new ObjectId()).firstName("Maximilian").deleted(true).build();
			User user3 = UserTestFactory.createBuilder().id(new ObjectId()).firstName("Maximus").deleted(false).build();

			repository.persist(user1);
			repository.persist(user2);
			repository.persist(user3);

			var foundUsers = repository.findUsers("Max", 10).toList();

			assertThat(foundUsers).hasSizeGreaterThan(2);
			List<User> startWithMax = foundUsers.stream().filter(u -> u.getFirstName().startsWith("Max")).toList();
			assertThat(startWithMax).hasSize(3);
			assertThat(startWithMax.stream().filter(User::isDeleted).count()).isEqualTo(2);
			assertThat(startWithMax.stream().filter(not(User::isDeleted)).count()).isEqualTo(1);
		}

		@Test
		void shouldNotFindAnyUser() {
			repository.persist(UserTestFactory.createBuilder().id(null).firstName("Marvel").build());

			var foundUsers = repository.findUsers("Max", 10).toList();

			assertThat(foundUsers).isEmpty();
		}

		@Test
		void shouldFindOneUser() {
			User user1 = UserTestFactory.createBuilder().id(null).firstName("Maximilian").build();
			User user2 = UserTestFactory.createBuilder().id(null).firstName("Madleen").build();

			repository.persist(user1);
			repository.persist(user2);

			var foundUsers = repository.findUsers("Max", 10).toList();

			assertThat(foundUsers)
					.hasSize(1)
					.first()
					.extracting(User::getFirstName)
					.isEqualTo("Maximilian");
		}
	}

	@Nested
	class TestFindByFullName {

		@BeforeEach
		void init() {
			repository.deleteAll();
		}

		@ParameterizedTest
		@MethodSource("provideDataForFindUsersByFullName")
		void shouldFindUsersByFullName(String query, String[] expectedUsernames) {
			Arrays.asList(
					UserTestFactory.createBuilder().id(null).username("user1").fullName("Franz Vogel").build(),
					UserTestFactory.createBuilder().id(null).username("user2").fullName("Franz von Holzhausen").build(),
					UserTestFactory.createBuilder().id(null).username("user3").fullName("Peter Lustig").build()).forEach(repository::persist);

			var foundUsernames = repository.findUsers(query, 10).map(User::getUsername);

			assertThat(foundUsernames).containsExactlyInAnyOrder(expectedUsernames);
		}

		private static Stream<Arguments> provideDataForFindUsersByFullName() {
			return Stream.of(
					Arguments.of("Fra", new String[] { "user1", "user2" }),
					Arguments.of("Franz", new String[] { "user1", "user2" }),
					Arguments.of("Franz v", new String[] { "user1", "user2" }),
					Arguments.of("Franz Vo ", new String[0]),
					Arguments.of("Franz von", new String[] { "user2" }),
					Arguments.of("Franz von Holzhausen", new String[] { "user2" }),
					Arguments.of("Franz L", new String[0]),
					Arguments.of("Peter V", new String[0]));
		}
	}

	@Nested
	class TestFindByFullNameReversed {

		@BeforeEach
		void init() {
			repository.deleteAll();
		}

		@ParameterizedTest
		@MethodSource("provideDataForFindUsersByFullNameReversed")
		void shouldFindUsersByFullNameReversed(String query, String[] expectedUsernames) {
			Arrays.asList(
					UserTestFactory.createBuilder().id(null).username("user1").fullNameReversed("Langbein Franz").build(),
					UserTestFactory.createBuilder().id(null).username("user2").fullNameReversed("Lustig Peter").build(),
					UserTestFactory.createBuilder().id(null).username("user3").fullNameReversed("Ilona Nowak").build()).forEach(repository::persist);

			var foundUsernames = repository.findUsers(query, 10).map(User::getUsername);

			assertThat(foundUsernames).containsExactlyInAnyOrder(expectedUsernames);
		}

		private static Stream<Arguments> provideDataForFindUsersByFullNameReversed() {
			return Stream.of(
					Arguments.of("L", new String[] { "user1", "user2", "user3" }),
					Arguments.of("Lustig", new String[] { "user2" }),
					Arguments.of("Lustig Peter", new String[] { "user2" }));
		}
	}

	@Nested
	class TestFindUsersByDeletedAndOrganisationsEinheitId {

		private static final String ORGANISATIONSEINHEIT_ID_1 = "ORGA_1";
		private static final String ORGANISATIONSEINHEIT_ID_2 = "ORGA_2";
		private static final int LIMIT = 10;
		private User user1;
		private User user2;
		private User user3;

		@BeforeEach
		void init() {
			user1 = UserTestFactory.createBuilder().id(null).build();
			user2 = UserTestFactory.createBuilder()
					.id(null)
					.clearOrganisationsEinheitIds()
					.organisationsEinheitId(ORGANISATIONSEINHEIT_ID_1)
					.build();
			user3 = UserTestFactory.createBuilder()
					.id(null)
					.clearOrganisationsEinheitIds()
					.organisationsEinheitId(ORGANISATIONSEINHEIT_ID_1)
					.build();
			repository.deleteAll();
		}

		@Test
		void shouldFindUserWithOrganisationsEinheitId() {
			persistTestUsers();

			var foundUsers = repository.findUsersByDeletedAndOrganisationsEinheitId(
					UserTestFactory.USER_NAME,
					false,
					UserTestFactory.ORGANISTATIONSEINHEITEN_ID,
					LIMIT).toList();

			assertThat(foundUsers)
					.hasSize(1)
					.first()
					.usingRecursiveComparison()
					.isEqualTo(user1);

		}

		@Test
		void shouldFindUsersWithOrganisationsEinheitId() {
			persistTestUsers();

			var foundUsers = repository.findUsersByDeletedAndOrganisationsEinheitId(
					UserTestFactory.USER_NAME,
					false,
					ORGANISATIONSEINHEIT_ID_1,
					LIMIT).toList();

			assertThat(foundUsers)
					.hasSize(2)
					.usingRecursiveFieldByFieldElementComparator()
					.contains(user2, user3);

		}

		@Test
		void shouldNotFindUserWithOrganisationsEinheitId() {
			persistTestUsers();

			var foundUsers = repository.findUsersByDeletedAndOrganisationsEinheitId(
					UserTestFactory.USER_NAME,
					false,
					ORGANISATIONSEINHEIT_ID_2,
					LIMIT).toList();
			assertThat(foundUsers).isEmpty();
		}

		private void persistTestUsers() {
			repository.persist(user1);
			repository.persist(user2);
			repository.persist(user3);
		}
	}

	@DisplayName("Find all inactive user")
	@Nested
	class TestFindAllInactiveUser {

		@BeforeEach
		void init() {
			repository.deleteAll();
			User user1 = UserTestFactory.createBuilder().id(null).firstName("Max").deleted(true).build();
			User user2 = UserTestFactory.createBuilder().id(null).firstName("Maximilian").deleted(true).build();
			User user3 = UserTestFactory.createBuilder().id(null).firstName("Maximus").deleted(false).build();
			repository.persist(user1);
			repository.persist(user2);
			repository.persist(user3);
		}

		@Test
		void shouldFindDeletedUsers() {
			var foundUsers = repository.findUsersByDeleted("Max", true, 10).toList();

			assertThat(foundUsers).hasSize(2);
			foundUsers.forEach(u -> assertThat(u.isDeleted()).isTrue());
		}

	}

	@DisplayName("Find by externalId")
	@Nested
	class TestFindByExternalId {

		@BeforeEach
		void init() {
			repository.deleteAll();
			repository.persist(UserTestFactory.create());
		}

		@Test
		void shouldFindByExternalId() {
			var user = repository.findByExternalId(UserTestFactory.EXTERNAL_ID);

			assertThat(user).isPresent();
			assertThat(user.get()).usingRecursiveComparison().isEqualTo(UserTestFactory.create());
		}
	}

	@Nested
	class TestFindByKeycloadId {

		@BeforeEach
		void init() {
			repository.deleteAll();
		}

		@Test
		void shouldFindUser() {
			repository.persist(UserTestFactory.create());

			var user = repository.findByKeycloakId(UserTestFactory.KEYCLOAK_USER_ID);

			assertThat(user)
					.isPresent()
					.map(User::getKeycloakUserId)
					.hasValue(UserTestFactory.KEYCLOAK_USER_ID);
		}

		@Test
		void shouldNotFindUser() {
			var user = repository.findByKeycloakId(UserTestFactory.KEYCLOAK_USER_ID);

			assertThat(user).isEmpty();
		}
	}

	@DisplayName("Find by id")
	@Nested
	class TestFindById {

		@BeforeEach
		void init() {
			repository.deleteAll();
			repository.persist(UserTestFactory.create());
		}

		@Test
		void shouldFindById() {
			var user = repository.findById(UserTestFactory.ID_STR);

			assertThat(user).isNotNull();
			assertThat(user.get()).usingRecursiveComparison().isEqualTo(UserTestFactory.create());
		}

		@Test
		void shouldNotFindIfIdNull() {
			assertThat(repository.findById((String) null)).isEmpty();
		}

		@Test
		void shouldNotFindIfEmptyId() {
			assertThat(repository.findById("")).isEmpty();
		}

		@Test
		void shouldNotFindIfBlankId() {
			assertThat(repository.findById(" ")).isEmpty();
		}

		@Test
		void shouldNotFindIfInvalidId() {
			assertThat(repository.findById("wrong_id")).isEmpty();
		}
	}

	@DisplayName("Find by email")
	@Nested
	class TestFindByEmail {

		@BeforeEach
		void init() {
			repository.deleteAll();
			repository.persist(UserTestFactory.create());
		}

		@Test
		void shouldFindByEmailOnly() {
			var user = repository.findByEmail(UserTestFactory.EMAIL);

			assertThat(user).isPresent();
			assertThat(user.get()).usingRecursiveComparison().isEqualTo(UserTestFactory.create());
		}

		@Test
		void shouldNotFindByEmptyEmail() {
			var user = repository.findByEmail(null);

			assertThat(user).isEmpty();
		}

	}

	@DisplayName("Test refreshing user")
	@Nested
	class TestRefresh {

		@BeforeEach
		void init() {
			repository.deleteAll();
			repository.persist(UserTestFactory.create());
		}

		@Test
		void shouldRefreshByExternalId() {
			var user = repository.refresh(UserTestFactory.createBuilder().id(null).build());

			assertThat(user).usingRecursiveComparison().ignoringFields("id").isEqualTo(UserTestFactory.create());
		}

		@Test
		void shouldRefreshById() {
			var user = repository.refresh(UserTestFactory.create());

			assertThat(user).usingRecursiveComparison().ignoringFields("id").isEqualTo(UserTestFactory.create());
		}
	}

	@Nested
	class TestFindAllDeletedUserIds {

		@BeforeEach
		void init() {
			repository.deleteAll();
		}

		@Test
		void shouldFindOneElement() {
			var deletedUser = UserTestFactory.createBuilder().id(null).deleted(true).build();
			repository.persist(deletedUser);

			var userIds = repository.findAllDeletedUserIds();

			assertThat(userIds).hasSize(1).first().isEqualTo(deletedUser.getId().toHexString());
		}

		@Test
		void shouldFindNoneElements() {
			repository.persist(UserTestFactory.create());

			var userIds = repository.findAllDeletedUserIds();

			assertThat(userIds).isEmpty();
		}
	}

	@Nested
	class TestFindAllActiveByOrganisationsEinheitId {

		private static final String ORGANISATIONSEINHEIT_ID_1 = UUID.randomUUID().toString();
		private static final String ORGANISATIONSEINHEIT_ID_2 = UUID.randomUUID().toString();

		private final User user1 = UserTestFactory.createBuilder()
				.id(null)
				.clearOrganisationsEinheitIds()
				.organisationsEinheitId(ORGANISATIONSEINHEIT_ID_1)
				.build();
		private final User user2 = UserTestFactory.createBuilder()
				.id(null)
				.clearOrganisationsEinheitIds()
				.organisationsEinheitId(ORGANISATIONSEINHEIT_ID_1)
				.build();
		private final User user3 = UserTestFactory.createBuilder()
				.id(null)
				.clearOrganisationsEinheitIds()
				.organisationsEinheitId(ORGANISATIONSEINHEIT_ID_2)
				.build();
		private final User deletedUser = UserTestFactory.createBuilder()
				.id(null)
				.clearOrganisationsEinheitIds()
				.organisationsEinheitId(ORGANISATIONSEINHEIT_ID_1)
				.deleted(true)
				.build();

		@BeforeEach
		void init() {
			repository.deleteAll();
			repository.persist(user1, user2, user3, deletedUser);
		}

		@Test
		void shouldFindActiveUsersByOrganisationsEinheitId() {
			var users = repository.findAllActiveByOrganisationsEinheitId(ORGANISATIONSEINHEIT_ID_1);

			assertThat(users).usingRecursiveFieldByFieldElementComparator().containsExactlyInAnyOrder(user1, user2);
		}
	}

	@Nested
	class TestDeleteById {

		@BeforeEach
		void init() {
			repository.deleteAll();
			repository.persist(UserTestFactory.create());
		}

		@Test
		void shouldDelete() {
			repository.deleteById(UserTestFactory.ID_STR);

			assertThat(repository.findAll().stream()).isEmpty();
		}
	}

}